﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using System.Reflection;
using System.Text;

namespace TeachingLinq
{
    public class DataDump
    {
        /// <summary>
        /// Representation of a simple object.
        /// </summary>
        private string representation;
        /// <summary>
        /// Representation of a recursive object.
        /// </summary>
        private List<DataDump> components;
        /// <summary>
        /// Depth of the data dump.
        /// </summary>
        private int Depth;
        /// <summary>
        /// Prefix of the representation.
        /// </summary>
        private string prefix;
        /// <summary>
        /// True if the representation is a string.
        /// </summary>
        private bool isString;

        /// <summary>
        /// Create a DataDump of an object o: a nice representation of a recursive tree-like object.
        /// </summary>
        /// <param name="o">Object to dump.</param>
        public DataDump(object o)
        {
            this.representation = null;
            this.components = null;
            this.prefix = "";
            this.isString = false;

            var t = o.GetType();
            if (t == typeof (string))
            {
                this.representation = o.ToString();
                this.Depth = 0;
                this.isString = true;
                return;
            }
            
            if (t.IsValueType)
            {
                this.representation = o.ToString();
                this.Depth = 0;
                return;
            }

            if (ImplementsInterface(t, typeof(IGrouping<,>)))
            {
                var key = GetField(o, "Key");

                this.prefix = key + " => ";
                this.components = DumpIenumerable(o, ref this.Depth);
                return;
            }
            
            if (t.IsArray)
            {
                this.components = DumpIenumerable(o, ref this.Depth);
                return;
            }
            
            if (ImplementsInterface(t, typeof(IEnumerable<>)))
            {
                this.components = DumpIenumerable(o, ref this.Depth);
                return;
            }

            // object
            this.components = DumpClass(o, ref this.Depth);
        }

        private static List<DataDump> DumpClass(object o, ref int depth)
        {
            List<DataDump> parts = new List<DataDump>();

            foreach (PropertyInfo p in o.GetType().GetProperties())
            {
                object value = p.GetValue(o, null);
                DataDump part = new DataDump(value);
                part.prefix = p.Name + " => ";
                if (part.Depth + 1 > depth)
                    depth = part.Depth + 1;
                parts.Add(part);
            }

            foreach (var f in o.GetType().GetFields())
            {
                object value = f.GetValue(o);
                DataDump part = new DataDump(value);
                part.prefix = f.Name + " => ";
                if (part.Depth + 1 > depth)
                    depth = part.Depth + 1;
                parts.Add(part);
            }

            return parts;
        }

        private static List<DataDump> DumpIenumerable(object o, ref int depth)
        {
            List<DataDump> parts = new List<DataDump>();
            IEnumerable ie = (IEnumerable)o;

            foreach (var c in ie)
            {
                DataDump part = new DataDump(c);
                if (part.Depth + 1 > depth)
                    depth = part.Depth + 1;
                parts.Add(part);
            }

            return parts;
        }

        private int EstimateLength()
        {
            if (this.representation != null)
            {
                return this.representation.Length;
            }
            else if (this.components != null)
            {
                return this.components.Select(c => c.EstimateLength()).Sum();
            }
            else
            {
                throw new InvalidOperationException();
            }
        }

        private string ToString(string spaces)
        {
            if (this.representation != null)
            {
                if (this.isString)
                    return spaces + this.prefix + "\"" + this.representation.Replace("\"", "\\\"") + "\"";
                else
                    return spaces + this.prefix + this.representation;
            }
            else
            {
                if (this.Depth >= 2 || this.EstimateLength() > 80)
                {
                    string innerspaces = "  " + spaces;
                    StringBuilder result = new StringBuilder();

                    result.Append(spaces);
                    if (!string.IsNullOrEmpty(this.prefix))
                    {
                        result.AppendLine(this.prefix);
                    }

                    result.AppendLine("{");
                    int length = this.components.Count();
                    foreach (var c in this.components)
                    {
                        result.Append(c.ToString(innerspaces));
                        if (--length > 0)
                            result.Append(",");
                        result.AppendLine();
                    }
                    result.Append(spaces + "}"); 
                    return result.ToString();
                }
                else
                {
                    // depth == 1, estimatedLength < 80, single line
                    bool first = true;
                    StringBuilder result = new StringBuilder();

                    result.Append(spaces);
                    if (!string.IsNullOrEmpty(this.prefix))
                    {
                        result.Append(this.prefix);
                    }

                    result.Append("{");
                    foreach (var c in this.components)
                    {
                        if (!first)
                            result.Append(",");
                        first = false;
                        result.Append(c.ToString(""));
                    }
                    result.Append("}"); 
                    return result.ToString();
                }
            }
        }

        /// <summary>
        /// String representation of the DataDump.
        /// </summary>
        /// <returns>A string describing the data dump.</returns>
        public override string ToString()
        {
            return this.ToString("");
        }

        private static bool ImplementsInterface(Type t, Type interf)
        {
            foreach (var intf in t.GetInterfaces())
            {
                if (intf == interf)
                    return true;
                if (intf.IsGenericType && intf.GetGenericTypeDefinition() == interf)
                    return true;
            }
            return false;
        }

        private static object GetField(object o, string fieldName)
        {
            var t = o.GetType();
            foreach (var f in t.GetFields())
            {
                if (f.Name == fieldName)
                    return f.GetValue(o);
            }
            foreach (var f in t.GetProperties())
            {
                if (f.Name == fieldName)
                    return f.GetValue(o, null);
            }
            return null;
        }
    }
}